#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#elif defined(WIN32)
#include <glut.h>
#else
#include <GL/glut.h>
#endif

#include "mesh.h"

int init(void);
void disp(void);
void idle(void);
void draw_text(float r, float g, float b, const char *fmt, ...);
void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);
void mouse(int bn, int state, int x, int y);
void motion(int x, int y);

float cam_theta = 125.0, cam_phi = 25.0, cam_dist = 6.0;
int win_width, win_height;

struct mesh *mesh;

int fps;					/* framerate */
unsigned int prev_fps_upd;	/* when we last updated the framerate */
unsigned int frames;		/* number of frames drawn since last fps update */

int main(int argc, char **argv)
{
	glutInit(&argc, argv);
	glutInitWindowSize(800, 600);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("foo");

	glutDisplayFunc(disp);
	glutIdleFunc(idle);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);

	if(init() == -1) {
		return 1;
	}

	glutMainLoop();
	return 0;
}


int init(void)
{
	glClearColor(0.1, 0.1, 0.1, 1);

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	/*
	glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
	*/

	if(!(mesh = load_mesh("statues.obj"))) {
		return -1;
	}

	return 0;
}

void disp(void)
{
	float diffuse[] = {0.2, 0.3, 1.0, 1.0};
	float specular[] = {1, 1, 1, 1};
	float lpos[] = {-30, 40, 50, 1};
	unsigned int msec = glutGet(GLUT_ELAPSED_TIME);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	/* set the light in view space, before the view transform, to position it
	 * in relation to the camera instead of the world
	 */
	glLightfv(GL_LIGHT0, GL_POSITION, lpos);

	glTranslatef(0, 0, -cam_dist);
	glRotatef(cam_phi, 1, 0, 0);
	glRotatef(cam_theta, 0, 1, 0);

	/* set material and render the mesh */
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, diffuse);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 60.0);
	render_mesh(mesh);

	/* show the fps counter text */
	draw_text(1, 1, 0, "fps: %d", fps);

	glutSwapBuffers();
	assert(glGetError() == GL_NO_ERROR);

	/* if a second has passed since the last fps update, count the frames
	 * and reset the time interval
	 */
	if(msec - prev_fps_upd >= 1000) {
		fps = frames;
		frames = 0;
		prev_fps_upd = msec;
	}
	frames++;
}

void idle(void)
{
	glutPostRedisplay();
}

void draw_text(float r, float g, float b, const char *fmt, ...)
{
	char buf[256], *text = buf;
	va_list ap;

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, win_width, 0, win_height, -1, 1);

	glDisable(GL_LIGHTING);

	glRasterPos2f(1, 1);
	glColor3f(r, g, b);
	while(*text) {
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *text++);
	}

	glEnable(GL_LIGHTING);

	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

void reshape(int x, int y)
{
	win_width = x;
	win_height = y;

	glViewport(0, 0, x, y);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, (float)x / (float)y, 0.5, 500);
}

void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
	case 'q':
	case 'Q':
		exit(0);

	case 'w':
	case 'W':
		{
			static int wire;

			wire = !wire;
			if(wire) {
				glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			} else {
				glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
			}
		}
		glutPostRedisplay();
		break;
	}
}

static int bnstate[16];
static int prev_x, prev_y;

void mouse(int bn, int state, int x, int y)
{
	int idx = bn - GLUT_LEFT_BUTTON;

	if(idx < 16) {
		bnstate[idx] = state == GLUT_DOWN;
	}

	prev_x = x;
	prev_y = y;
}


void motion(int x, int y)
{
	int dx = x - prev_x;
	int dy = y - prev_y;
	prev_x = x;
	prev_y = y;

	if(bnstate[0]) {
		cam_theta += dx * 0.5;
		cam_phi += dy * 0.5;
		if(cam_phi < -90) cam_phi = -90;
		if(cam_phi > 90) cam_phi = 90;
		glutPostRedisplay();
	}

	if(bnstate[2]) {
		cam_dist += dy * 0.1;
		if(cam_dist < 0.0) {
			cam_dist = 0.0;
		}
		glutPostRedisplay();
	}
}
